
// MarvelUtils.java
// Andrew Davison, February 2014, ad@fivedots.coe.psu.ac.th

/*
   Support methods for building clients that access the
   the Marvel Rest API (http://developer.marvel.com/)

  The methods here are divided into groups:

     * methods for processing/parsing the user's command line input
       (e.g. converting strings into year ranges)

     * methods for constructing the URL strings, suitable for querying the
       Marvel Rest API. These can be determined by trying out interactive queries
       at the developers site (http://developer.marvel.com/docs#!/public) if
       you have an account.

     * network methods for connecting to the portal using the URL, and reading
       the response as a string

     * methods for parsing the response tring as a JSON structure, and extracting
       the necessary information.

     * comic image downloading, display, and saving methods
*/

import java.awt.*;
import javax.swing.*;
import java.awt.event.*;
import java.io.*;
import java.net.*;
import java.security.*;
import java.util.*;
import javax.imageio.*;
import java.awt.image.*;

import org.json.simple.*;
import org.json.simple.parser.*;
   // for JSON parsing

import org.apache.commons.codec.binary.*;     
   // required for the Hex encoding used by in genMD5hash()


public class MarvelUtils
{

  private static final String MARVEL_ADDR = "http://gateway.marvel.com/v1/public/";
      // address of Marvel's developers portal



  // ========================= processing of user's command line input =========


  public static int[] getYears(String yearRangeStr)
  // convert a "year" (or "year-year") string into an array of integers
  {
    int[] yearRange = new int[2];
    yearRange[0] = -1;    // from-year
    yearRange[1] = -1;    // to-year

    int yearSeperator = yearRangeStr.indexOf("-");
    if (yearSeperator == -1) { // extract one year
      int fromYear = toInt(yearRangeStr);
      if (fromYear == -1)
        return null;
      else
        yearRange[0] = fromYear;
    }
    else {   // extract two years
      String y1Str = yearRangeStr.substring(0, yearSeperator);
      String y2Str = yearRangeStr.substring(yearSeperator+1);

      // extract the from-year
      int fromYear = toInt(y1Str);
      if (fromYear == -1) {
        System.out.println("No from-year found");
        return null;
      }
      else
        System.out.println("Found from-year: " + fromYear);

      yearRange[0] = fromYear;
      
      // extract the to-year
      int toYear = toInt(y2Str);
      if (toYear == -1) 
        System.out.println("No to-year found; using only the from-year");
      else
        System.out.println("Found to-year: " + toYear);

      yearRange[1] = toYear;    // may be -1
    }
    return yearRange;
  }  // end of getYears()



  public static int toInt(String str)
  // parse a string into an integer
  {
    int val = -1;
    try {
      val = Integer.parseInt(str);
    }
    catch(NumberFormatException e)
    {  System.out.println(str + " not an integer"); }
    return val;
  }  // end of toInt()



  public static boolean isID(String s)
  /* check if a string is a Marvel ID (i.e. a sequence of digit chars)
     and possibly a comma, if the string hold multiple IDs
     e.g. "1987,454"
  */
  {
    if (s == null)
      return false;

    char ch;
    for (int i = 0; i < s.length(); i++) {
      ch = s.charAt(i);
      if (!Character.isDigit(ch) && (ch != ','))
        return false;
    }
    return true;
  }   // end of isID()



  public static int[] getIDs(String s)
  // convert IDs string into an array of ID integers
  {
    if (!isID(s))
      return null;
    else {
      String[] idStrs = s.split(",");
      int[] ids = new int[idStrs.length];
      for (int i=0; i < idStrs.length; i++)
        ids[i] = toInt(idStrs[i]);
      return ids;
    }
  }  // end of getIDs()



  // ============= construct query  and process response ====================

  /* shows the basic 3-step use of most REST APIs, including Marvel's: 
        - construct the URL string
        - use the URL string to connect to the Rest API; read the response string
        - parse the response, extracting info
  */


  public static String lookupCharacterID(String characterName, 
                                        String publicKey, String privateKey)
  // used in GetID.java: get IDs for a character name;
  // return first matching character ID
  {
    String nameQuery = characterRequest(characterName, publicKey, privateKey);
    String nameResponse = sendQuery(nameQuery);
    return characterResults(nameResponse);
  }  // end of lookupCharacterID()



  public static String lookupSeriesID(String seriesName, 
                                        String publicKey, String privateKey)
  // used in GetID.java: get IDs for a series name;
  // return first matching series ID
  {
    String seriesQuery = seriesRequest(seriesName, publicKey, privateKey);
    String seriesResponse = sendQuery(seriesQuery);
    return seriesResults(seriesResponse);
  }   // end of lookupSeriesID()



  // ======================== construct query =======================



  public static String characterRequest(String name, 
                                        String publicKey, String privateKey)
  // search the "characters" table for a given (partial) character name;
  // uses "nameStartsWith" to increase chances of a match
  {
    Long timeStamp = new Date().getTime();
    String hash = genMD5hash(timeStamp + privateKey + publicKey);
              // according to http://developer.marvel.com/documentation/authorization

    String urlStr = MARVEL_ADDR + "characters?" +   // URL to access the characters table
        "limit=50" +
        "&nameStartsWith=" + URLencode(name) + 
        "&ts=" + timeStamp + 
        "&apikey=" + publicKey + 
        "&hash=" + hash;

    return urlStr;
  }  // end of characterRequest()



  private static String genMD5hash(String msg)
  // transforms message string to a MD5 hash string
  {
    String hash = null;
    try {
      MessageDigest md = MessageDigest.getInstance("MD5");

      byte[] digestBytes = md.digest( msg.getBytes("UTF-8") );

      hash = new String( Hex.encodeHex(digestBytes) );
                 // Hex class is from org.apache.commons.codec.binary
    }
    catch (NoSuchAlgorithmException e) {
      System.out.println("No MD5 algorithm found");
      System.exit(1);
    }
    catch (UnsupportedEncodingException e) {
      System.out.println("No Hex encoder found");
      System.exit(1);
    }
    return hash;
  }  // end of genMD5hash()


  private static String URLencode(String msg)
  // encode URI reserved characters, such as '/' and ' ';
  // for details, see http://en.wikipedia.org/wiki/Percent-encoding
  {
    String msgEncode = null;
    try {
      msgEncode = URLEncoder.encode(msg, "UTF-8");
    }
    catch(UnsupportedEncodingException e) {
      System.out.println("Could not encode " + msg);
      msgEncode = msg;
    }
    return msgEncode;
  }  // end of URLencode()




  public static String seriesRequest(String seriesName, 
                                        String publicKey, String privateKey)
  /* search the "series" table for a given (partial) series name;
     uses "titleStartsWith" to increase chances of a match;
     return the series in start year order 
  */
  {
    Long timeStamp = new Date().getTime();
    String hash = genMD5hash(timeStamp + privateKey + publicKey);

    String urlStr = MARVEL_ADDR + "series?" +    // URL to access the series table
        "limit=50" +
        "&titleStartsWith=" + URLencode(seriesName) +
        "&orderBy=startYear" +
        "&ts=" + timeStamp + 
        "&apikey=" + publicKey + 
        "&hash=" + hash;

    return urlStr;
  }  // end of seriesRequest()



  public static String characterIDRequest(String characterID, 
                                        String publicKey, String privateKey)
  // search the "characters" table for a given character ID;
  {
    Long timeStamp = new Date().getTime();
    String hash = genMD5hash(timeStamp + privateKey + publicKey);
              // according to http://developer.marvel.com/documentation/authorization

    String urlStr = MARVEL_ADDR + "characters/" + characterID + "?" +
        "&ts=" + timeStamp + 
        "&apikey=" + publicKey + 
        "&hash=" + hash;

    return urlStr;
  }  // end of characterIDRequest()




  public static String seriesIDRequest(String seriesID, 
                                        String publicKey, String privateKey)
  // search the "series" table for a given series ID;
  {
    Long timeStamp = new Date().getTime();
    String hash = genMD5hash(timeStamp + privateKey + publicKey);

    String urlStr = MARVEL_ADDR + "series/" + seriesID + "?" +
        "&ts=" + timeStamp + 
        "&apikey=" + publicKey + 
        "&hash=" + hash;

    return urlStr;
  }  // end of seriesIDRequest()



  public static String characterSeriesRequest(String characterID, 
                                        String publicKey, String privateKey)
  /* search the "characters" table for series using the given character ID;
     return the series in start year order */
  {
    Long timeStamp = new Date().getTime();
    String hash = genMD5hash(timeStamp + privateKey + publicKey);

    String urlStr = MARVEL_ADDR + "characters/" + characterID + "/series?" + 
        "limit=50" +
        "&orderBy=startYear" +
        "&ts=" + timeStamp + 
        "&apikey=" + publicKey + 
        "&hash=" + hash;

    return urlStr;
  }  // end of characterSeriesRequest()




  public static String seriesCharactersRequest(String seriesID, 
                                        String publicKey, String privateKey)
  // search the "series" table for characters appearing in the given series ID
  {
    Long timeStamp = new Date().getTime(); 
    String hash = genMD5hash(timeStamp + privateKey + publicKey);

    String urlStr = MARVEL_ADDR + "series/" + seriesID + "/characters?" +
        "limit=50" +
        "&ts=" + timeStamp + 
        "&apikey=" + publicKey + 
        "&hash=" + hash;

    return urlStr;
  }  // end of seriesCharactersRequest()



  public static String comicsRequest(String charId, String seriesId, int[] yearRange,
                                     String publicKey, String privateKey)
  /* search the "comics" table for a given character ID (or IDs), series ID 
     (or IDs) and a possible year range,
     returning the matching comics ordered by on-sale date.

     Each ID string may contain multiple IDs, separated by a comma with no
     extra spaces. If there are multiple character IDs, then look for a shared
     appearance.
*/
  {
    if ((charId == null) && (seriesId == null) && (yearRange == null)) {
      System.out.printf("Can not construct a comics query with no data");
      return null;
    }

    int[] ids = getIDs(charId);
    boolean isShared = false;
    if ((ids != null) && (ids.length > 1)) {
      System.out.println("Looking for a shared appearance of IDs: " + charId);
      isShared = true;
    }

    Long timeStamp = new Date().getTime();
    String hash = genMD5hash(timeStamp + privateKey + publicKey);

    String urlStr = MARVEL_ADDR + "comics?" +
        "limit=50" +
        ((charId == null) ? "" : (
           (isShared ? ("&sharedAppearances=" + URLencode(charId)) :
                       ("&characters=" + URLencode(charId))) 
         )) +
        ((seriesId == null) ? "" : ("&series=" + URLencode(seriesId)) ) +
        ((yearRange == null) ? "" : ("&dateRange=" + buildDateRange(yearRange)) ) +
        "&orderBy=onsaleDate" +
        "&ts=" + timeStamp + 
        "&apikey=" + publicKey + 
        "&hash=" + hash;

    return urlStr;
  }  // end of comicsRequest()




  private static String buildDateRange(int[] yearRange)
  /* The year range are a start year and end year for a date-based search.
     Convert the two years into dates by adding "01-01" and "-12-31" to
     the start and end year strings respectively.

     The year range array may only contain a single year, so use that as the
     date range. The array mey be null, so use this year.

     The two years may be in the wrong order, so switch them.
  */
  {
    int currYear = Calendar.getInstance().get(Calendar.YEAR);

    if (yearRange == null) {    // no year info supplied; use current year
      System.out.println("Using current year (" + currYear + ") by default");
      return (currYear + "-01-01" + "%2C" + currYear + "-12-31");    // %2C = ','
    }
    else if (yearRange[1] == -1)  // only one year supplied
      return (yearRange[0] + "-01-01" + "%2C" + yearRange[0] + "-12-31");
    else {   // two years supplied
      if (yearRange[0] > yearRange[1]) {   // years are in wrong order, so switch
        System.out.println("Years being swapped so range makes sense"); 
        return (yearRange[1] + "-01-01" + "%2C" + yearRange[0] + "-12-31"); 
      }
      else
        return (yearRange[0] + "-01-01" + "%2C" + yearRange[1] + "-12-31"); 
    }
  }  // end of buildDateRange()





  public static String comicIDRequest(String id,
                           String publicKey, String privateKey)
  // search the "comics" table for a specific comic ID
  {
    Long timeStamp = new Date().getTime();
    String hash = genMD5hash(timeStamp + privateKey + publicKey);

    String urlStr = MARVEL_ADDR + "comics/" + id + "?" +
        // "limit=50" +
        "&ts=" + timeStamp + 
        "&apikey=" + publicKey + 
        "&hash=" + hash;

    return urlStr;
  }  // end of comicIDRequest()



  // ======================= networking connection =======================


  public static String sendQuery(String urlStr)
  // contact the specified URL, and return the response as a string
  {
    String response = null;
    HttpURLConnection conn = null;
    System.out.println("Query URL: " + urlStr);
    try {
      URL url = new URL(urlStr);

      conn = (HttpURLConnection) url.openConnection();
      conn.setRequestMethod("GET");
      conn.setRequestProperty("Accept", "application/json");

      if (conn.getResponseCode() != 200) {
        System.out.println("Failed; HTTP error code: " + conn.getResponseCode() + 
                            " " + conn.getResponseMessage());
        return null;
      }

      BufferedReader br = new BufferedReader(
                 new InputStreamReader((conn.getInputStream())));
      String line = null;
      StringBuilder sb = new StringBuilder();

      while ((line = br.readLine()) != null)
        sb.append(line);
      br.close();

      response = sb.toString();
    }
    catch (MalformedURLException e) {
      System.out.println("Error: Not a valid URL");
    } 
    catch (IOException e) {
      System.out.println("Error: Could not retrieve search results");
    }
    finally {
      conn.disconnect();
    }
    return response;
  }  // end of sendQuery()



  // ============================ process response =====================



  public static String characterResults(String queryResponse)
  /* For each JSONObject in the response array for characters, 
     print the name, id, 
     description, and thumbnail fields. The description and
     thumbnail forelds may not be present.

     Return the first ID found in the list.
  */
  {
    JSONArray results = extractResults(queryResponse);
    if (results == null)
      return null;

    String id = null;
    String firstID = null;
    try{
      if (results.size() != 1)
        System.out.println("Size of results array: " + results.size());
      for(int i=0; i < results.size(); i++) {
        JSONObject result = (JSONObject)results.get(i);

        id = "" + result.get("id");
        if (i == 0)
          firstID = id;
        System.out.println((i+1) + ". name: \"" + result.get("name") + "\"; ID: " + id);

        String description = (String)result.get("description");
        if ((description != null) && (description.length() > 0))
          System.out.println("  description: \"" + description + "\"");

        JSONObject thumbnail = (JSONObject)result.get("thumbnail");
        if (thumbnail != null)
          System.out.println("  thumbnail: \"" + thumbnail.get("path") + "." + 
                                         thumbnail.get("extension") + "\"");

        System.out.println();
      }
    }
    catch(Exception e)
    {  System.out.println(e);  }
    return firstID;
  }  // end of characterResults()




  private static JSONArray extractResults(String jsonText)
  /* convert response string into a JSON array. The format
     of the response string is:
         {
           "code": 200,
           "status": "Ok",
           "etag": "....",
           "data": {
             "offset": 0,
             "limit": 20,
             "total": 1,
             "count": 1,
  		       "results": [ {...} {...} {...} ... ]
          }}

     Dig down to the "results" field inside "data", and return it.
      Also report the "status" field value and the "total" value.
  */
  {
    if (jsonText == null)
      return null;

    JSONArray results = null;
    JSONParser parser = new JSONParser();
    try{
      JSONObject jo = (JSONObject)parser.parse(jsonText);
           // response string becomes a JSON object

      // look inside the JSON object for the results array
      String status = ((String)jo.get("status")).toLowerCase();
      System.out.println("\nResult status: " + status);
      if (!status.equals("ok"))
        return null;

      JSONObject jod = (JSONObject)jo.get("data");

      System.out.println("Total matches: " + jod.get("total"));
      results = (JSONArray)jod.get("results");
    }
    catch(Exception e)
    {  System.out.println("Could not extract results: " + e);  }

    return results;
  } // end of extractResults()




  public static String seriesResults(String queryResponse)
  /* For each JSONObject in the response array for series, 
     print the title, id, 
     and description. The description field may not be present.

     Return the first ID found in the list.
  */
  {
    JSONArray results = extractResults(queryResponse);
    if (results == null)
      return null;

    String id = null;
    String firstID = null;
    try{
      if (results.size() != 1)
        System.out.println("Size of results array: " + results.size());
      for(int i=0; i < results.size(); i++) {
        JSONObject result = (JSONObject)results.get(i);

        id = "" + result.get("id");
        if (i == 0)
          firstID = id;
        System.out.println((i+1) + ". title: \"" + result.get("title") + "\"; ID: " + id);

        String description = (String)result.get("description");
        if ((description != null) && (description.length() > 0))
          System.out.println("  description: \"" + description + "\"");
/*
        JSONObject thumbnail = (JSONObject)result.get("thumbnail");
        if (thumbnail != null)
          System.out.println("  thumbnail: \"" + thumbnail.get("path") + "." + 
                                         thumbnail.get("extension") + "\"");
*/
        System.out.println();
      }
    }
    catch(Exception e)
    {  System.out.println(e);  }
    return firstID;
  }  // end of seriesResults()




  public static String comicsResults(String queryResponse, boolean showFull)
  /* For each JSONObject in the response array for comics, 
     print the title, issue no., on-sale date, description, thumbnail, id,
     page count, and story details. If the showFull flag is true then also
     include creator and character info.

     Return the first thumbnail URL address string found in the list.
  */
  {
    String thumbAddress = null;
    JSONArray results = extractResults(queryResponse);
    if (results == null)
      return null;

    try{
      if (results.size() != 1)
        System.out.println("Size of results array: " + results.size());
      for(int i=0; i < results.size(); i++) {
        JSONObject result = (JSONObject)results.get(i);

        System.out.println((i+1) + ". Title: \"" + result.get("title") + "\";  issue no: " + 
                                          result.get("issueNumber"));
        reportSaleDate(result);

        String description = (String)result.get("description");
        if ((description != null) && (description.length() > 0))
          System.out.println("  description: \"" + description + "\"");


        JSONObject thumbnail = (JSONObject)result.get("thumbnail");
        if (thumbnail != null) {
          thumbAddress = thumbnail.get("path") + "." + thumbnail.get("extension");
          System.out.println("  thumbnail: \"" + thumbAddress + "\"");
        }

        System.out.println("  id: " + result.get("id"));
        System.out.println("  page count: " + result.get("pageCount"));

        reportStories( (JSONObject)result.get("stories") );

        if (showFull) {
          reportCreators( (JSONObject)result.get("creators") );
          reportCharacters( (JSONObject)result.get("characters") );
        }

        System.out.println();
      }

    }
    catch(Exception e)
    {  System.out.println(e);  }

    return thumbAddress;
  }  // end of comicsResults()




  private static void reportSaleDate(JSONObject result)
  /*  Print the on-sale Date for the comic. JSON format:
        dates": [
          {
            "type": "onsaleDate",
            "date": "1965-09-10T00:00:00-0400"
          },,  ... ]
  */
  {
    JSONArray dates = (JSONArray)result.get("dates");
    // System.out.println("  No. of dates: " + dates.size());
         // there may be many dates, stored in a JSON array
    for(int j=0; j < dates.size(); j++) {    // print only the on-sale date, if present
      JSONObject date = (JSONObject)dates.get(j);
      if (((String)date.get("type")).equals("onsaleDate")) {   // when it went on sale
        String dateStr = (String)date.get("date");
        System.out.println("  On-sale date: " + shortenDate(dateStr));
        break;
      }
    }
  }  // end of reportSaleDate()



  private static String shortenDate(String date)
  // strip out time part of the date string, so only the date is reported;
  // e.g. remove everything from 'T' in "1965-09-10T00:00:00-0400"
  {
    int timePos = date.indexOf("T");
    if (timePos == -1)  // not found
      return date;
    else
      return date.substring(0, timePos);
  }  // end of shortenDate()




  private static void reportStories(JSONObject stories)
  /* Print all the story titles in the given comic.
     The titles are stored in a JSON array called items, and includes an entry
     for the cover whose 'title' is a number.
     JSON format:
     items": [
            {
              "resourceURI": "http://gateway.marvel.com/v1/public/stories/34249",
              "name": "Cover #34249",
              "type": "cover"
            },  ... ]
  */
  {
    JSONArray items = (JSONArray)stories.get("items");
    System.out.println("  No. of stories: " + items.size());
    for(int i=0; i < items.size(); i++) {
      JSONObject item = (JSONObject)items.get(i);
      if (((String)item.get("type")).equals("cover"))    // the cover title
        System.out.println("    cover: \"" + item.get("name") + "\"");
      else
        System.out.println("    story: \"" + item.get("name") + "\"");
    }
  }  // end of reportStories()



  private static void reportCreators(JSONObject creators)
  /* Print all the creators involved in the given comic.
     The information is stored in a JSON array called items, and includes a
     creator name and their role.
     JSON format:
         items": [
            {
              "resourceURI": "http://gateway.marvel.com/v1/public/creators/1523",
              "name": "Vince Colletta",
              "role": "inker"
            },  ... ]
  */
  {
    JSONArray items = (JSONArray)creators.get("items");
    System.out.println("  No. of creators: " + items.size());
    for(int i=0; i < items.size(); i++) {
      JSONObject item = (JSONObject)items.get(i);
        System.out.println("    " + item.get("role") + ": " + item.get("name"));
    }
  }  // end of reportCreators()




  private static void reportCharacters(JSONObject characters)
  /* Print all the characters mentioned in the given comic.
     The information is stored in a JSON array called items, and includes the
     character name.
     JSON format:
          items": [
            {
              "resourceURI": "http://gateway.marvel.com/v1/public/characters/1009664",
              "name": "Thor"
            },  ... ]
  */
  {
    JSONArray items = (JSONArray)characters.get("items");
    System.out.println("  No. of characters: " + items.size());
    for(int i=0; i < items.size(); i++) {
      JSONObject item = (JSONObject)items.get(i);
        System.out.println("    " + item.get("name"));
    }
  }  // end of reportCharacters()




  // =========================== comic image utils ======================


  public static BufferedImage downloadImage(String urlStr)
  { 
    BufferedImage im = null;

    // Read from a URL
    System.out.println("Downloading \"" + urlStr + "\" ...");
    try {
      im = ImageIO.read( new URL(urlStr) );
    }
    catch(IOException e)
    {  System.out.println("Could not download URL: " + e);  }

    return im;
  }  // end of downloadImage()



  public static void displayImage(BufferedImage im, String imFnm)
  /* load the image as a label inside a JFrame, including a menu
     bar for saving the image if the user wants 
  */
  {
    JFrame f = new JFrame("Cover Image");
    f.getContentPane().add( new JLabel(new ImageIcon(im)) );
    makeMenuBar(f, im, imFnm);

    f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    f.pack();
    f.setResizable(false);
    f.setLocationRelativeTo(null);  // center the window 
    f.setVisible(true);
  }  // end of displayImage()




  private static void makeMenuBar(JFrame f, 
                        final BufferedImage im, final String imFnm)
  /* the frame's menu bar has a single menu item for
     saving the image into the specified file
  */
  {
    JMenuBar menubar = new JMenuBar();
    f.setJMenuBar(menubar);
        
    // create the File menu
    JMenu fileMenu = new JMenu("File");
    menubar.add(fileMenu);
        
    JMenuItem saveItem = new JMenuItem("Save to " + imFnm);
    saveItem.addActionListener( new ActionListener(){
      public void actionPerformed(ActionEvent e)
      {   saveImage(im, imFnm);  }
    });
    fileMenu.add(saveItem);
  }  // end of makeMenuBar()




  public static void saveImage(BufferedImage im, String fnm)
  {
    System.out.println("Saving downloaded image to " + fnm);
    try {
      ImageIO.write(im, "png", new File(fnm));
    }
    catch(IOException e)
    {  System.out.println("Could not save: " + e);  }
  }  // end of saveImage()


}  // end of MarvelUtils class
